#include <stdlib.h>
#include <string.h>
#include "matrix.h"

#define ptr_check(a,b)    \
	if (a == NULL) {  \
		return b; \
	}

static int _matrix_add_matrix(matrix *res, const matrix *a, const matrix *b, short mode);
static int _matrix_num(matrix *mx, double num, short mode);

matrix
*matrix_new()
{
	matrix *mx;

	mx = malloc(sizeof(*mx));
	if (mx != NULL) {
		mx->rows = 0;
		mx->cols = 0;
		mx->val = NULL;
	}

	return mx;
}

int
matrix_free(matrix **mx)
{
	ptr_check(*mx, MATRIX_USERR)
	ptr_check(mx, MATRIX_UNSET)

	if ((*mx)->val != NULL) {
		free((*mx)->val);
	}

	free(*mx);
	(*mx) = NULL;

	return MATRIX_DONE;
}

int
matrix_clear(matrix *mx)
{
	ptr_check(mx, MATRIX_UNSET)
	ptr_check(mx->val, MATRIX_NOVAL)

	ptr_check(memset(mx->val, 0, mx->rows * mx->cols * sizeof(double)), MATRIX_NOVAL)

	return MATRIX_DONE;
}

int
matrix_resize(matrix *mx, unsigned short rows, unsigned short cols)
{
	double *res;

	ptr_check(mx, MATRIX_UNSET)

	if (mx->val == NULL) {
		mx->val = malloc(rows * cols * sizeof(double));
		ptr_check(mx->val, MATRIX_NOMEM)
	}

	res = realloc(mx->val, rows * cols * sizeof(double));
	ptr_check(res, MATRIX_NOMEM)

	mx->rows = rows;
	mx->cols = cols;

	return (matrix_clear(mx));
}

int
matrix_set(matrix *mx, const double *val, unsigned short rows, unsigned short cols)
{
	int res;

	ptr_check(mx, MATRIX_UNSET)
	ptr_check(val, MATRIX_NOVAL)

	res = matrix_resize(mx, rows, cols);
	if (res != MATRIX_DONE)	{
		return res;
	}

	ptr_check(memcpy(mx->val, val, rows * cols * sizeof(*val)), MATRIX_NOVAL)

	return MATRIX_DONE;
}

int
matrix_get_element(double *dst, const matrix *mx, unsigned short row, unsigned short col)
{
	ptr_check(dst, MATRIX_USERR)
	ptr_check(mx, MATRIX_UNSET)
	ptr_check(mx->val, MATRIX_NOVAL)

	if (row >= mx->rows) {
		return MATRIX_USERR;
	}

	if (col >= mx->cols) {
		return MATRIX_USERR;
	}

	*dst = mx->val[row * mx->cols + col];

	return MATRIX_DONE;
}

int
matrix_set_element(double val, matrix *mx, unsigned short row, unsigned short col)
{
	ptr_check(mx, MATRIX_UNSET)
	ptr_check(mx->val, MATRIX_NOVAL)

	if (row >= mx->rows) {
		return MATRIX_USERR;
	}

	if (col >= mx->cols) {
		return MATRIX_USERR;
	}

	mx->val[row * mx->cols + col] = val;

	return MATRIX_DONE;
}

int
matrix_copy(matrix *dst, const matrix *src)
{
	int res;

	ptr_check(src, MATRIX_UNSET)

	if (dst == NULL) {
		dst = matrix_new();
		ptr_check(dst, MATRIX_NOMEM)
	}

	res = matrix_set(dst, src->val, src->rows, src->cols);

	return res;
}

int
matrix_rsign(matrix *mx)
{
	int i;

	ptr_check(mx, MATRIX_UNSET)
	ptr_check(mx->val, MATRIX_NOVAL)

	for (i = 0; i < mx->rows * mx->cols; i++) {
		mx->val[i] = -mx->val[i]; 
	}

	return MATRIX_DONE;
}

static int
_matrix_add_matrix(matrix *res, const matrix *a, const matrix *b, short mode)
{
	matrix *tmp;
	int i;

	ptr_check(a, MATRIX_UNSET)
	ptr_check(b, MATRIX_UNSET)
	ptr_check(a->val, MATRIX_NOVAL)
	ptr_check(b->val, MATRIX_NOVAL)

	if (a->rows != b->rows) {
		return MATRIX_USERR;
	}

	if (a->cols != b->cols) {
		return MATRIX_USERR;
	}

	tmp = matrix_new();
	ptr_check(tmp, MATRIX_NOMEM)
	i = matrix_resize(tmp, a->rows, b->cols);
	if (i != MATRIX_DONE) {
		return i;
	}

	if (mode == 0) {
		for (i = 0; i < tmp->rows * tmp->cols; i++) {
			tmp->val[i] = a->val[i] + b->val[i];
		}
	}

	if (mode == 1) {
		for (i = 0; i < tmp->rows * tmp->cols; i++) {
			tmp->val[i] = a->val[i] - b->val[i];
		}
	}



	i = matrix_copy(res, tmp);
	matrix_free(&tmp);

	return i;
}

int
matrix_add_matrix(matrix *res, const matrix *a, const matrix *b)
{
	return _matrix_add_matrix(res, a, b, 0);
}

int
matrix_sub_matrix(matrix *res, const matrix *a, const matrix *b)
{
	return _matrix_add_matrix(res, a, b, 1);
}

int
matrix_mult_matrix(matrix *res, const matrix *a, const matrix *b)
{
	matrix *tmp1, *tmp2;
	double buf;
	int i, j, k;

	tmp1 = matrix_new();
	ptr_check(tmp1, MATRIX_NOMEM)

	tmp2 = matrix_new();
	ptr_check(tmp2, MATRIX_NOMEM)

	i = matrix_copy(tmp1, a);
	if (i != MATRIX_DONE) {
		return i;
	}

	i = matrix_copy(tmp2, b);
	if (i != MATRIX_DONE) {
		return i;
	}

	if (tmp1->cols == tmp2->cols) {
		if (matrix_transpose(tmp2) != MATRIX_DONE) {
			return i;
		}
	}

	if (tmp1->rows == tmp2->rows) {
		if (matrix_transpose(tmp1) != MATRIX_DONE) {
			return i;
		}
	}

	if (tmp1->cols != tmp2->rows) {
		return MATRIX_USERR;
	}

	i = matrix_resize(res, tmp1->rows, tmp2->cols);
	if (i != MATRIX_DONE) {
		return i;
	}

	for (i = 0; i < tmp1->rows; i++) {
		for (j = 0; j < tmp2->cols; j++) {
			buf = 0;
			for (k = 0; k < tmp2->rows; k++) {
				 buf += tmp1->val[i * tmp1->cols + k] * tmp2->val[tmp2->rows * k + j];
				 printf("|%i:%i:%i| ", i, j, k);
			}
			res->val[i * res->cols + j] = buf;
			printf("\n");
		}
	}

	matrix_free(&tmp1);
	matrix_free(&tmp2);

	return MATRIX_DONE;
}

static int
_matrix_num(matrix *mx, double num, short mode)
{
	int i;

	ptr_check(mx, MATRIX_UNSET)
	ptr_check(mx->val, MATRIX_NOVAL)

	if (mode == 0) {
		for (i = 0; i < mx->rows * mx->cols; i++) {
			mx->val[i] += num;
		}
	}

	if (mode == 1) {
		for (i = 0; i < mx->rows * mx->cols; i++) {
			mx->val[i] *= num;
		}
	}

	return MATRIX_DONE;
}

int
matrix_add_number(matrix *mx, double num)
{
	return _matrix_num(mx, num, 0);
}

int
matrix_mult_number(matrix *mx, double num)
{
	return _matrix_num(mx, num, 1);
}

int
matrix_transpose(matrix *mx)
{
	double *tmp;
	int i, j;

	ptr_check(mx, MATRIX_UNSET)
	ptr_check(mx->val, MATRIX_NOVAL)

	tmp = malloc(mx->rows * mx->cols * sizeof(double));

	for (i = 0; i < mx->rows; i++) {
		for (j = 0; j < mx->cols; j++) {
			tmp[mx->rows * j + i] = mx->val[mx->cols * i + j];
		}
	}

	i = matrix_set(mx, tmp, mx->cols, mx->rows);

	return i;
}
